// tutorial03.c
// A pedagogical video player that will stream through every video frame as fast as it can
// and play audio (out of sync).
//
// Code based on FFplay, Copyright (c) 2003 Fabrice Bellard,
// and a tutorial by Martin Bohme (boehme@inb.uni-luebeckREMOVETHIS.de)
// Tested on Gentoo, CVS version 5/01/07 compiled with GCC 4.1.1
// Use
//
// gcc -o tutorial03 tutorial03.c -lavformat -lavcodec -lz -lm `sdl-config --cflags --libs`
// to build (assuming libavformat and libavcodec are correctly installed,
// and assuming you have sdl-config. Please refer to SDL docs for your installation.)
//
// Run using
// tutorial03 myvideofile.mpg
//
// to play the stream on your screen.

#include <libswscale/swscale.h>
#include <libavcodec/avcodec.h>
#include <libavformat/avformat.h>

#include <SDL/SDL.h>
#include <SDL/SDL_thread.h>

#ifdef __MINGW32__
#undef main /* Prevents SDL from overriding main() */
#endif

#include <stdio.h>

//#define SDL_AUDIO_BUFFER_SIZE 1024
#define SDL_AUDIO_BUFFER_SIZE AVCODEC_MAX_AUDIO_FRAME_SIZE

typedef struct PacketQueue
{
    AVPacketList *first_pkt, *last_pkt;
    int nb_packets;
    int size;
    SDL_mutex *mutex;
    SDL_cond *cond;
} PacketQueue;

PacketQueue audioq;

int quit = 0;

void img_convert(AVPicture* pDst, enum PixelFormat dstFmt, AVPicture* pSrc, enum PixelFormat srcFmt, int srcW, int srcH)
{
    struct SwsContext* context = NULL; // =
    context = sws_getContext(srcW, srcH, srcFmt, srcW, srcH, dstFmt, SWS_BILINEAR, NULL, NULL, NULL);
    //context = sws_alloc_context();
    //sws_init_context(context, NULL, NULL);
    sws_scale(context, pSrc->data, pSrc->linesize, 0, srcH, pDst->data, pDst->linesize);
    //sws_freeContext(context);
}

void packet_queue_init(PacketQueue *q)
{
    memset(q, 0, sizeof(PacketQueue));
    q->mutex = SDL_CreateMutex();
    q->cond = SDL_CreateCond();
}

int packet_queue_put(PacketQueue *q, AVPacket *pkt)
{
    AVPacketList *pkt1;
    if (av_dup_packet(pkt) < 0)
    {
        return -1;
    }
    pkt1 = av_malloc(sizeof(AVPacketList));
    if (!pkt1)
        return -1;
    pkt1->pkt = *pkt;
    pkt1->next = NULL;

    SDL_mutexP(q->mutex);

    if (!q->last_pkt)
        q->first_pkt = pkt1;
    else
        q->last_pkt->next = pkt1;
    q->last_pkt = pkt1;
    q->nb_packets++;
    q->size += pkt1->pkt.size;
    SDL_CondSignal(q->cond);
    SDL_mutexV(q->mutex);
    return 0;
}

static int packet_queue_get(PacketQueue *q, AVPacket *pkt, int block)
{
    AVPacketList *pkt1;
    int ret;

    SDL_LockMutex(q->mutex);

    for (;;)
    {
        if (quit)
        {
            ret = -1;
            break;
        }

        pkt1 = q->first_pkt;
        if (pkt1)
        {
            q->first_pkt = pkt1->next;
            if (!q->first_pkt)
                q->last_pkt = NULL;
            q->nb_packets--;
            q->size -= pkt1->pkt.size;
            *pkt = pkt1->pkt;
            av_free(pkt1);
            ret = 1;
            break;
        }
        else if (!block)
        {
            ret = 0;
            break;
        }
        else
        {
            SDL_CondWait(q->cond, q->mutex);
        }
    }
    SDL_UnlockMutex(q->mutex);
    return ret;
}

int audio_decode_frame(AVCodecContext *aCodecCtx, uint8_t *audio_buf, int buf_size)
{
    static AVPacket pkt;
    // this is to make sure pointer alignment requirement of api
    // and the buffer size is bit too big, (static to avoid stack size overflow)
    static int16_t data_buffer[AVCODEC_MAX_AUDIO_FRAME_SIZE];
    // no static to ensure init every call !
    int write_size = 0;
    int read_size = 0;

    if (packet_queue_get(&audioq, &pkt, 1) <= 0)
    {
        return -1;
    }
    else
    {
        // get packet
        if (pkt.data)
        {
            write_size = AVCODEC_MAX_AUDIO_FRAME_SIZE*sizeof(int16_t);
            read_size = avcodec_decode_audio3(aCodecCtx, data_buffer, &write_size, &pkt);
            if (read_size > 0 && write_size > 0 && write_size <= buf_size)
            {
                memcpy(audio_buf, data_buffer, write_size);
            }
            else
            {
                write_size = 0;
            }
            av_free_packet(&pkt);
        }
    }
    return write_size;
}


void audio_callback(void *userdata, Uint8 *stream, int len)
{
    static uint8_t audio_buf[AVCODEC_MAX_AUDIO_FRAME_SIZE * 3 ];
    static uint8_t* ptrStart = audio_buf;
    static uint8_t * ptrCurrent = audio_buf;
    static int data_size = 0;
    AVCodecContext *aCodecCtx = (AVCodecContext *)userdata;
    int read_size;

    if (len > 0)
    {
        if (data_size < len && data_size > 0)
        {
            // let's move data to beginning of buffer and restart reading data into buffer
            memmove(audio_buf, ptrStart + data_size, data_size);
            ptrStart = audio_buf;
            ptrCurrent = ptrStart + data_size;
        }
        while (data_size < len)
        {
            read_size = audio_decode_frame(aCodecCtx, ptrCurrent, AVCODEC_MAX_AUDIO_FRAME_SIZE);
            if (read_size <= 0)
            {
                /* If error, output silence */
                memset(ptrCurrent, 0, len - data_size);
                data_size = len;
            }
            else
            {
                data_size += read_size;
            }
            ptrCurrent = ptrStart + data_size;
        }
        memcpy(stream, ptrStart, len);
        data_size -= len;
        if (data_size > 0)
        {
            ptrStart += len;
        }
        else
        {
            ptrStart = audio_buf;
        }

        ptrCurrent = ptrStart + data_size;
    }
}


int main(int argc, char *argv[])
{
    AVFormatContext *pFormatCtx;
    int             i, videoStream, audioStream;
    AVCodecContext  *pCodecCtx;
    AVCodec         *pCodec;
    AVFrame         *pFrame;
    AVPacket        packet;
    int             frameFinished;
    float           aspect_ratio;

    AVCodecContext  *aCodecCtx;
    AVCodec         *aCodec;

    SDL_Overlay     *bmp;
    SDL_Surface     *screen;
    SDL_Rect        rect;
    SDL_Event       event;
    SDL_AudioSpec   wanted_spec, spec;

    if (argc < 2)
    {
        fprintf(stderr, "Usage: test <file>\n");
        exit(1);
    }
    // Register all formats and codecs
    av_register_all();

    if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_TIMER))
    {
        fprintf(stderr, "Could not initialize SDL - %s\n", SDL_GetError());
        exit(1);
    }

    // Open video file
    if (av_open_input_file(&pFormatCtx, argv[1], NULL, 0, NULL)!=0)
        return -1; // Couldn't open file

    // Retrieve stream information
    if (av_find_stream_info(pFormatCtx)<0)
        return -1; // Couldn't find stream information

    // Dump information about file onto standard error
    dump_format(pFormatCtx, 0, argv[1], 0);

    // Find the first video stream
    videoStream=-1;
    audioStream=-1;
    for (i=0; i<pFormatCtx->nb_streams; i++)
    {
        if (pFormatCtx->streams[i]->codec->codec_type==CODEC_TYPE_VIDEO &&
                videoStream < 0)
        {
            videoStream=i;
        }
        if (pFormatCtx->streams[i]->codec->codec_type==CODEC_TYPE_AUDIO &&
                audioStream < 0)
        {
            audioStream=i;
        }
    }
    if (videoStream==-1)
        return -1; // Didn't find a video stream
    if (audioStream==-1)
        return -1;

    aCodecCtx=pFormatCtx->streams[audioStream]->codec;
    // Set audio settings from codec info
    wanted_spec.freq = aCodecCtx->sample_rate;
    wanted_spec.format = AUDIO_S16SYS;
    wanted_spec.channels = aCodecCtx->channels;
    wanted_spec.silence = 0;
    wanted_spec.samples = SDL_AUDIO_BUFFER_SIZE;
    wanted_spec.callback = audio_callback;
    wanted_spec.userdata = aCodecCtx;

    if (SDL_OpenAudio(&wanted_spec, &spec) < 0)
    {
        fprintf(stderr, "SDL_OpenAudio: %s\n", SDL_GetError());
        return -1;
    }
    aCodec = avcodec_find_decoder(aCodecCtx->codec_id);
    if (!aCodec)
    {
        fprintf(stderr, "Unsupported codec!\n");
        return -1;
    }
    avcodec_open(aCodecCtx, aCodec);

    // audio_st = pFormatCtx->streams[index]
    packet_queue_init(&audioq);
    SDL_PauseAudio(0);

    // Get a pointer to the codec context for the video stream
    pCodecCtx=pFormatCtx->streams[videoStream]->codec;

    // Find the decoder for the video stream
    pCodec=avcodec_find_decoder(pCodecCtx->codec_id);
    if (pCodec==NULL)
    {
        fprintf(stderr, "Unsupported codec!\n");
        return -1; // Codec not found
    }
    // Open codec
    if (avcodec_open(pCodecCtx, pCodec)<0)
        return -1; // Could not open codec

    // Allocate video frame
    pFrame=avcodec_alloc_frame();

    // Make a screen to put our video

#ifndef __DARWIN__
    screen = SDL_SetVideoMode(pCodecCtx->width, pCodecCtx->height, 0, 0);
#else
    screen = SDL_SetVideoMode(pCodecCtx->width, pCodecCtx->height, 24, 0);
#endif
    if (!screen)
    {
        fprintf(stderr, "SDL: could not set video mode - exiting\n");
        exit(1);
    }

    // Allocate a place to put our YUV image on that screen
    bmp = SDL_CreateYUVOverlay(pCodecCtx->width,
                               pCodecCtx->height,
                               SDL_YV12_OVERLAY,
                               screen);

    // Read frames and save first five frames to disk
    i=0;
    while ((i = av_read_frame(pFormatCtx, &packet))>=0)
    {
        // Is this a packet from the video stream?
        if (packet.stream_index==videoStream)
        {
            // Decode video frame

            avcodec_decode_video(pCodecCtx, pFrame, &frameFinished, packet.data, packet.size);

            // Did we get a video frame?
            if (frameFinished)
            {
                SDL_LockYUVOverlay(bmp);

                AVPicture pict;
                pict.data[0] = bmp->pixels[0];
                pict.data[1] = bmp->pixels[2];
                pict.data[2] = bmp->pixels[1];

                pict.linesize[0] = bmp->pitches[0];
                pict.linesize[1] = bmp->pitches[2];
                pict.linesize[2] = bmp->pitches[1];

                // Convert the image into YUV format that SDL uses
                img_convert(&pict, PIX_FMT_YUV420P,
                            (AVPicture *)pFrame, pCodecCtx->pix_fmt,
                            pCodecCtx->width, pCodecCtx->height);

                SDL_UnlockYUVOverlay(bmp);

                rect.x = 0;
                rect.y = 0;
                rect.w = pCodecCtx->width;
                rect.h = pCodecCtx->height;
                SDL_DisplayYUVOverlay(bmp, &rect);
                av_free_packet(&packet);
            }

            av_free_packet(&packet);
        }
        else if (packet.stream_index==audioStream)
        {
            packet_queue_put(&audioq, &packet);
        }
        else
        {
            av_free_packet(&packet);
        }
        // Free the packet that was allocated by av_read_frame
        SDL_PollEvent(&event);
        switch (event.type)
        {
        case SDL_QUIT:
            quit = 1;
            SDL_Quit();
            exit(0);
            break;
        default:
            break;
        }

    }

    // Free the YUV frame
    av_free(pFrame);

    // Close the codec
    avcodec_close(pCodecCtx);

    // Close the video file
    av_close_input_file(pFormatCtx);

    return 0;
}
